#################################################################################
# Created       : Martin Foster
# Modified      : 25-Nov-2005
#################################################################################
#
# Login - Modules part of Ethereal Realms, designed to unify authenthication 
# Copyright (C) 2000-2004  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 29-26503 TWP RD 511
#                 Spruce Grove, Alberta
#                 T7Y 1G4
#
#################################################################################

package Ethereal::Login;					# Package name

use Ethereal::Audit;						# Audit trail
use Ethereal::Template;						# Template handler

#################################################################################
# Sub-Routines
#################################################################################

#####################
# Constructor
#
# Allows to class to be exported and thus spawn objects as a result

sub new
{
	#####################
	# Data members

        my $class    = shift;					# Reference to self
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter hash
	my $template = shift;					# Optional template hash

	my $self;						# Self referncing link


	#####################
	# Program area

	# Define parameters
        $self  = {
		CGI      => $cgi,
		HANDLE   => $database,
		PARAM    => $param, 
		TEMPL    => $template,
	       	HEADER   => undef,
		FOOTER   => undef,
		AUTOMATE => 1,
		GENERATE => 0,
		NODEPEND => 1,
		DONE     => 0
	 };

        # Bless $self into designated class
        bless $self, $class;

        # Return self
        return $self;
}



#####################
# Authenthication
#
# This is the primary component of this module, and this would be the 
# authenthication script, which will verify usernames and passwords.

sub Authenthication
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $choice   = shift;					# Choice of the matter

	my $audit;						# Auditing entry
	my $cgi      = $self->{'CGI'};				# Common gateway interface
	my $database = $self->{'HANDLE'};			# Database handle
	my $param    = $self->{'PARAM'};			# Parameter list

	my $check;						# Check registration	
	my $crypt;						# Crypted password
	my $curr;						# Current address
	my $query;						# Query to ask
	my $user;						# Users login name

	my $encode;						# Encoded string
	my $mori;						# Kill switch
	
	my $cid;						# Cookie ID
	my $recall;						# Recall cookie
	my $secure;						# Security cookie
	my $span;						# Life span of cookie
	
	my %login;						# Login hash


	#####################
	# Program area

        # Session timeouts
	if (defined($param->{'MORI'}))
	{
		# Longer elapsed then it should be
		if ((time - $param->{'MORI'}) > 3600)
		{
			# Return false
			return 0;
		}

		# Cleanup parameters
		$param->Cleanup('MORI');
	}


	# Build query
	# Administrative
	if ($choice == 1)
	{ 
		# Restrict to administrators
		$query = "SELECT PuppeteerLogin, 
			  PuppeteerPassword
			FROM Puppeteer 
			WHERE LOWER(PuppeteerLogin)=LOWER(?)
			  AND PuppeteerAdminPrivs='yes'"; 
	}

	# Supervisory
	elsif ($choice == 2)
	{ 
		# Restrict to moderators and supervisors
		$query = "SELECT PuppeteerLogin, 
			  PuppeteerPassword 
			FROM Puppeteer 
			WHERE LOWER(PuppeteerLogin)=LOWER(?)
			  AND (PuppeteerSuperPrivs='yes'
			   OR PuppeteerModPrivs='yes')";
	}

	# Non-authoritative
	elsif (($choice >= 99) || ($choice == 3))
	{ 
		# General all allow
		$query = "SELECT PuppeteerLogin, 
			  PuppeteerPassword 
			FROM Puppeteer 
			WHERE LOWER(PuppeteerLogin)=LOWER(?)"; 
	}

	# Retrieve information
	($user, $crypt) = $database->DataGet($query, $param->{'USER'});


	# Compare passwords
	if (defined($crypt))
	{
		# If crypt
		if (defined($param->{'CRYPT'})) 
		{
			# Failed authenthication
		 	if ($param->{'CRYPT'} ne $crypt)
			{
				# Audit entry
				$audit = new Ethereal::Audit($database, $cgi, $param);
				$audit->Pass();

				# Authenthiation failed
				return 0; 
			}
		}

		# If first time
		elsif (defined($param->{'PASS'}))
		{
			# Fail on non-match
			if (crypt(substr($param->{'PASS'}, 0, 10), $crypt) ne $crypt)
			{
				# Audit entry
				$audit = new Ethereal::Audit($database, $cgi, $param);
				$audit->Pass();

				# Authenthication Failed
				return 0;
			}

			# Otherwise assign crypt
			else
			{ $param->{'CRYPT'} = (crypt(substr($param->{'PASS'}, 0, 10), $crypt)); }
		}

		# Assume fault
		else
		{ return 0; }
	}

	# Assume fault
	else
	{ return 0; }


	# Verify for incomplete registration
	($check) = $database->DataGetPuppeteerReg($user);

	# Return false if registration is incomplete
	if (defined($check))
	{ return 0; }


	# Database updates
	# Disabling of Javascript
	if (defined($param->{'DISABLE_JAVASCRIPT'}))
	{
		# Update database entry
		$database->Quick("UPDATE Puppeteer 
			SET   PuppeteerJavascript='no' 
			WHERE PuppeteerLogin=?", $user);
	}

	# New authenthication login
	if ((defined($param->{'PASS'}))
	 || (defined($param->{'COOKIE'})))
	{
		# Evaluation
		eval
		{
			# Record authenthication
			$audit = new Ethereal::Audit($database, $cgi, $param);
			$audit->Auth();


			# Generate cookies
			# More permenant cookie
			if ((defined($param->{'USER_RECALL'}))
			 || (defined($param->{'USER_FORGET'})))
			{
				# Kill switch after a month
				$mori = time + 6696000;

				# Encrypt parameters
				$encode = $param->EncodeString(
					USER  => $user, 
					CRYPT => $param->{'CRYPT'},
					MORI  => $mori
	 			);

				# Life span of cookie
				$span = (defined($param->{'USER_RECALL'})) ? '+1M' : '-1Y';
				
				# Generate cookies
				$recall = $cgi->cookie(
					 -NAME		=> 'EREALMS',
					 -VALUE		=> $encode,
					 -DOMAIN	=> '.' . $database->{'DOMAIN'},
					 -EXPIRES 	=> $span
					);

				# Push cookie
				push(@{$self->{'COOKIE'}}, $recall);

				# Send notice
				$self->{'GENERATE'} = 1;
			}

			# Session Cookie
			unless (defined($param->{'USER_SECURE'}))
			{
				# Handling of Cookie ID
				# Revert to defaults
				if (defined($param->{'XCKID'}))
				{
					# Re-Assign ID
					$cid = $param->{'XCKID'};
				}

				# Only if not already defined
				else
				{
					# Legacy random
					srand();

					# Generate ID
					$cid = int(rand() * 10000);

					# Begin changes in parameters
					$param->{'XCKID'} = $cid;

					# Generate cookies
					$secure = $cgi->cookie(
						 -NAME		=> "EREALMS\_$cid",
						 -VALUE		=> $user,
						 -DOMAIN	=> '.' . $database->{'DOMAIN'}
						);

					# Push cookie
					push(@{$self->{'COOKIE'}}, $secure);

					# Send notice
					$self->{'GENERATE'} = 1;
				}

				# Commit
				$database->Commit();
			}
		};

		# If block failed
		if ($@)
		{
			# Issue warning
			warn("Transaction aborted: $@");

			# Undo incomplete changes
			$database->Rollback;


			# Error header
			print $cgi->header(-status=>'503 Service Temporarily Unavailable');

			# Prevent dups
			$self->{'DONE'} = 1;

			
			# Return False
			return 0;
		}
	} 

	# Cleanup
	$param->Cleanup( 
		'COOKIE', 
		'PASS' 
	 );

	# Final assignmnet
	$param->{'USER'} = $user;


	# Return true
	return 1;
}


####################
# Check Legality
#
# Determins the legality of the user using the currently selected realm.  This 
# will also verify if a realm and puppet were selected.    

sub CheckLegality
{
	####################
	# Data members	
	my $self     = shift;					# Reference to self
	my $realm    = shift;					# Realm hash

	my $cgi      = $self->{'CGI'};				# Common gateway interface
	my $database = $self->{'HANDLE'};			# Database handle
	my $param    = $self->{'PARAM'};			# Parameter list

	my $tmpl;						# Template management

	my $access;						# Realm access
	my $pass;						# Realm password
	my $verdict;						# To be or not to be
	my $verified;						# Puppet/Realm verification 

	my $count;						# Counter
	my $message;						# Warning message
	my $redirect;						# Redirect information


	####################
	# Program area

	# Verification of needed information
	if (defined($param->{'ROOM'}))
	{
		# Verification of access
		# Realm
		($access, $verified) = $database->DataGetRealmAccess($param->{'ROOM'});
		($pass) = $database->DataGetRealmPassword($param->{'ROOM'});


		# Puppet default
		unless (defined($param->{'CHAR'}))
		{
			# Current handle retrieval
			($param->{'CHAR'}) = $database->DataGetPuppetCurrent(
				$param->{'ROOM'},
				$param->{'USER'}
			 );

			# Puppet default retreival
			($param->{'CHAR'}) = $database->DataGetDefault(
				$param->{'USER'}, 
				$param->{'ROOM'}, 
				$access
			 ) unless (defined($param->{'CHAR'}));
		}

		# If puppet is defined
		if (defined($param->{'CHAR'}))
		{
			# Users
			($verdict) = $database->DataGetBannedStatus($param->{'USER'}, $param->{'ROOM'});


			# Password verification
			if (defined($pass))
			{
				# Initialize counter
				$count = 0;

				# Ecrypted passwords
				if ((defined($param->{'RCRYPT'}))
				 && ($param->{'RCRYPT'} eq $pass))
				{
					# Increment
					$count++;
				}

				# Plain text
				elsif ((defined($param->{'RPASS'})) 
				 && (crypt($param->{'RPASS'}, $pass) eq $pass))
				{
					# Increment
					$count++;


					# Create encrypted counterpart
					$param->{'RCRYPT'} = crypt($param->{'RPASS'}, $pass);

					# Cleanup
					$param->Cleanup('RPASS');
				}

				# Invalid password
				else
				{
					# Assign message
					$message = $realm->{'MsgPassword'}; 
				}
			}
			else
			{
				# Initialize counter
				$count = 1;
			}


			# Verified realms
			# Realm verification of users
			if ($verified eq 'yes')
			{
				# Pull privilege
				($verified) = $database->DataGetVerified($param->{'USER'});

				# Verification
				# Verified user
				if ($verified eq 'yes')
				{
					# Increment for allowance
					$count++;
				}

				# Tsk tsk
				else
				{
					# Assign error message
					$message = $realm->{'MsgVerified'};
				}
			}
			else
			{
				# Assume success
				$count++;
			}

		
			# Access control
			# Relaxed/Restricted realms
			if (($access ne 'list')
			 && (!defined($verdict)))
			{ $count++; }

			# List based access/canon realms
			elsif (($access eq 'list')
			 && (defined($verdict)))
			{ $count++; }

			# Banned
			else
			{
		 		# Warn of being banned
				$message = $realm->{'MsgBanned'}; 
			}

	
			# Verify for complete check
			if ($count > 2)
			{
				# Return successful
				return 1;
			}
		}
	}

	# Default error message
	$message = (defined($message)) ? $message : $realm->{'MsgInvalid'};

	# Redirection
	# Retreive inline parameters
	$redirect = $param->EmbedInline(
		USER  =>$param->{'USER'},
		CRYPT =>$param->{'CRYPT'}
	 );


	# File header
	print $cgi->header(
		-type=>'text/html',
		-refresh=>"35;URL=$realm->{'LnkPortal'}$redirect"
	 );

	# Dirty deed complete
	$self->{'DONE'} = 1;


	# Template hash
	$tmpl     = new Ethereal::Template($realm);

	# HTML Display
	# Header
	$database->DocumentGetHeader();

	# Message Display
	$tmpl->Show('TmplRedirect', 
		MLINK => $realm->{'LnkPortal'},
		MMESS => $message
	 );

	# Footer
	$database->DocumentGetFooter();


	# Premature exit
	exit;
}


#####################
# GatherInfo
#
# Sub-routine that will handle the display of information, comes in two flavors
# adminstrative and non-authoritative login.   Selected though alias.

sub GatherInfo
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $choice   = shift;					# Choice on the matter

	my $cgi      = $self->{'CGI'};				# Common gateway interface
	my $database = $self->{'HANDLE'};			# Database handle
	my $param    = $self->{'PARAM'};			# Parameter list
	my $template = $self->{'TEMPL'};			# Alternate template hash
	my $header   = $self->{'HEADER'};			# Alternate header
	my $footer   = $self->{'FOOTER'};			# Alternate footer

	my $tmpl;						# Template handler

	my $ermes;						# Error message
	my $title;						# Title displayed
	my $scrpt;						# Javascript disabling
	my $session;						# Disabling of sessions
	
	my $wauth;						# Submit button
	my $wpass;						# Password field
	my $wrecl;						# Recall username and password
	my $wsecr;						# Insecure logins
	my $wuser;						# User field
	
	my %login;						# Login hash


	#####################
	# Program area

	# Link with hash
	$database->GetHashLogin(\%login);

	# Define Template instance
	$template = (defined($template)) ? $template : \%login;
	$tmpl     = new Ethereal::Template($template);


	# Determine header and footer
	$header = $database->DocumentGetHeader() unless (defined($header));
	$footer = $database->DocumentGetFooter() unless (defined($footer));


	# Cleanup
	$param->Cleanup('CRYPT','PASS','USER','MORI');


	# Populate and prepare
	# Select titles
	$title = ($choice > 2) ? $login{'TagAreaGeneral'} : $login{'TagAreaRestricted'};

	# Error display
	$ermes = (defined($param->{'AUTH'})) ? $login{'ErrLogin'} : '';
	
	# Javascript disable
	$scrpt = "<NOSCRIPT>" . $cgi->hidden('DISABLE_JAVASCRIPT','True') . "</NOSCRIPT>";


	# Generate widgets
	$wuser = $cgi->textfield('USER', '', 15, 10);
	$wpass = $cgi->password_field('PASS', '', 15, 15);
	$wauth = $cgi->submit($login{'TxtSubmitButton'});

	# Generate checkboxes
	$wrecl = $cgi->checkbox(-name=>'USER_RECALL', -label=>''); 
	$wsecr = $cgi->checkbox(-name=>'USER_SECURE', -label=>'');

	# HTML
	# Indepedant behaviour
	if ($self->{'NODEPEND'} != 0)
	{
		# Headers
		print $cgi->header();
		print $header, "\n";

		# Dirty deed complete
		$self->{'DONE'} = 1;
	}

	# Start from and embed
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat(), AUTH=>'True'), "\n";

	# Display template
	$tmpl->Show('TmplAuth',
		MERROR => $ermes,
		MTITLE => $title,
		MSCRPT => $scrpt,
		WAUTH  => $wauth,
		WPASS  => $wpass,
		WRECL  => $wrecl,
		WSECR  => $wsecr,
		WUSER  => $wuser
	 );

	# End of form
	print $cgi->end_form(), "\n";

	# Indepedant behaviour
	if ($self->{'NODEPEND'} != 0)
	{
		# If lacking information
		print $footer, "\n";
	}
}



#####################
# Get Verification
#
# Primary entrance which will lead to GatherInfo or Authenthication depending
# on received parameters.

sub GetVerification
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $choice   = shift;					# Choice on the matter

	my $cgi      = $self->{'CGI'};				# Common gateway interface
	my $database = $self->{'HANDLE'};			# Database handle
	my $param    = $self->{'PARAM'};			# Parameter list

	my $cookie;						# System cookie

	#####################
	# Program area

	# Verify is necessary components are there
	if ((defined($param->{'USER'}))
	 && ((defined($param->{'PASS'}))
	  || (defined($param->{'CRYPT'}))))
	{
		# Determine and return
		if ($self->Authenthication($choice))
		{
			# Generate header
			$self->Header();

			# Return true
			return 1;
		}
	}


	# Retrieve cookie
	$cookie = $cgi->cookie(-name=>'EREALMS');

	# Cookie exists
	if (defined($cookie))
	{
		# Retreive infromation
		$param->Cookie($cookie);

		# Repeat test
		if ((defined($param->{'USER'}))
		 && (defined($param->{'CRYPT'})))
		{
			# Assign check value
			$param->{'COOKIE'} = $param->{'USER'};

			# Verify authenthication
			if ($self->Authenthication($choice))
			{
				# Generate header
				$self->Header();

				# Return true
				return 1;
			}
		}
	}
	

	# Show login screen
	$self->GatherInfo($choice); 

	# Return false
	return 0;
}

# Alias for administrative
sub GetVerificationAdmin
{ return GetVerification(shift, 1); }

# Alias for normal
sub GetVerificationNormal
{ return GetVerification(shift, 3); }

# Alias for portal settings
sub GetVerificationPortal
{ return GetVerification(shift, 99); }

# Alias for realm specific
sub GetVerificationRealm
{ return GetVerification(shift, 100); }

# Alias for supervisory
sub GetVerificationSuper
{ return GetVerification(shift, 2); }


#####################
# Header
#
# Writes the file header, necessary for the proper display of HTML
# documents and the deployment of cookies.  This specific function
# is internal, the external version will have one value defaulted.

sub Header
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $manual   = shift;					# Manually create header
	my $redirect = shift;					# Redirection ability

	my $cgi      = $self->{'CGI'};				# CGI handle
	my $auto     = $self->{'AUTOMATE'};			# Automate header generation
	my $create   = $self->{'GENERATE'};			# Create cookies
	my $done     = $self->{'DONE'};				# Completed before
	
	my $header;						# Actual file header
	my $url;						# Self referencing link
	
	#####################
	# Program Area
	
	# Display header
	if ((($auto != 0)
	 or (defined($manual)))
         and ($done != 1))
	{
		# Variants of headers
		# Redirection header
		if (defined($redirect))
		{
			# Confirm redirect link is complete
			if ($redirect =~ /^http\:\/\//i)
			{
				# Pull base
				$url = $cgi->url(-base=>1);

				# Rewrite appropriatly 
				$redirect = $url . $redirect 
			}
			
			# Cookies are not supported
			$header = $cgi->redirect(-uri=>$redirect, @_);
		}

		# Regular header
		else
		{
			# One with or without cookies
			$header = ($create != 0)
			 ? $cgi->header(-cookie=>$self->{'COOKIE'}, @_)
			 : $cgi->header(@_);
		}

		# Establish done
		$self->{'DONE'} = 1;

		# Display
		print $header;
	}
}

# External Alias
sub WriteHeader
{ Header(shift, 1, @_); }

1;								# Return true
